/*
 * codegen.c - code generator
 * (c) 2000 Zeljko Vrba (mordor@fly.srk.fer.hr)
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <config.h>
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <stdarg.h>
#include "node.h"
#include "util.h"
#include "saturn.h"
#include "hpc-y.h"

#ifdef __GNUC__
static char rcsid[] __attribute__((unused)) = "$Id: codegen.c,v 1.1 2000-01-08 09:24:29+01 zeljko Exp $";
#else
static char rcsid[] = "$Id: codegen.c,v 1.1 2000-01-08 09:24:29+01 zeljko Exp $";
#endif

/*	Command line options. */
extern int opt_t, opt_o;
extern char *entrylib_path;

/*	Output stream. */
static FILE *out;

/*	Buffer... */
static char buf[1024];

/*	Loop environment counter for @0, @1 and break. */
static int loopenv[256], lstk;

/*	Forward declarations. */
static void compile_tree(struct st_node*);
static void output_symbols();
void optimize_code(struct st_node*);

/*	Front end to code generation. */
void output_code(const char *name, struct st_node *root)
{
	lstk = 0;
	if(opt_o) optimize_code(root);
	if(opt_t) {
		char buf[1024];

		sprintf(buf, "%s.s", name);
		if(!(out = fopen(buf, "w"))) {
			fprintf(stderr, "there was an error writing to output file; aborting\n");
			exit(1);
		}
	} else {
		saturn_openw(name, entrylib_path);
	}

	/* download header */
	if(opt_t) print(out, "ASSEMBLE\n\tNIBASC /HPHP48-R/\nRPL\n");

	/* code generation */
	print(out, "::");
	if(localsz > 0) output_symbols();
	indent_level = 1; nl(out);
	compile_tree(root); indent_level = 0; nl(out);
	if(localsz > 0) print(out, "ABND");
	indent_level = 0; nl(out);
	print(out, ";");
	if(opt_t) {
		fclose(out);
	} else {
		if(saturn_close() < 0) {
			fprintf(stderr, "there was an error writing to output file; aborting\n");
			exit(1);
		}
	}
}

/*	Outputs the code to BIND local variables and arguments to LAMs before
 *	the code for the subroutine is generated. */
static void output_symbols()
{
	int i;

	/*	Local variables get initialized to some equivalent of ZERO */
	for(i = 0; i < localsz; i++) {
		if(local_data[i]->scope == LOCAL) {
			switch(local_data[i]->type_tag) {
			case REAL:
				print(out, opt_t ? "%%0" : "%0");
				break;
			case LONGREAL:
				print(out, opt_t ? "%%%%0" : "%%0");
				break;
			case NATURAL:
				print(out, "ZERO");
				break;
			case STRING:
				print(out, "NULL$");
				break;
			case BINARY:
				print(out, "NULLHXS");
				break;
			case CHAR:
				print(out, "CHR_00");
				break;
			case COMPLEX:
				print(out, opt_t ? "C%%0" : "C%0");
				break;
			case LONGCOMPLEX:
				print(out, "C%%%% %f %f", 0., 0.);
				break;
			case LIST:
				print(out, "NULL{}");
				break;
			case ARRAY:
				if(local_data[i]->class.array.ndims == 1) {
					print(out, "{");
					print(out, "# %d", local_data[i]->class.array.dims[0]);
					print(out, "}");
				} else if(local_data[i]->class.array.ndims == 2) {
					print(out, "{");
					print(out, "# %d", local_data[i]->class.array.dims[0]);
					print(out, "# %d", local_data[i]->class.array.dims[1]);
					print(out, "}");
				} else {
					FATAL("invalid array dimensions");
				}
				if(local_data[i]->class.array.elemtype->type_tag == REAL) {
					print(out, opt_t ? "%%0" : "%0");
				} else if(local_data[i]->class.array.elemtype->type_tag == COMPLEX) {
					print(out, opt_t ? "C%%0" : "C%0");
				} else {
					FATAL("invalid array type");
				}
				print(out, "MAKEARRY");
				break;
			default:
				FATAL("invalid local variable type");
			}
		}
	}
	print(out, "{");
	for(i = 0; i < localsz; i++) {
		print(out, "LAM %s", local_names[i]);
	}
	print(out, "} BIND");
}

/*	Outputs an operator string depending on the type. */
static void oper(struct st_node *op)
{
	switch(op->type->type_tag) {
	case NATURAL:		strcpy(buf, "#"); break;
	case REAL:			strcpy(buf, opt_t ? "%%" : "%"); break;
	case LONGREAL:		strcpy(buf, opt_t ? "%%%%" : "%%"); break;
	case COMPLEX:		strcpy(buf, opt_t ? "C%%" : "C%"); break;
	case LONGCOMPLEX:	strcpy(buf, opt_t ? "C%%%%" : "C%%"); break;
	case BINARY:		strcpy(buf, "bit"); break;
	default:			error(op->line, "unsupported operator for type");
	}
	switch(op->opcode) {
	case '+': 	strcat(buf, "+"); break;
	case '-':	strcat(buf, "-"); break;
	case '*':	strcat(buf, "*"); break;
	case '/':
		strcat(buf, "/");
		if(op->type->type_tag == NATURAL) strcat(buf, " SWAPDROP");
		break;
	case '%':
		if(op->type->type_tag == REAL) {
			strcat(buf, "MOD");
		} else if(op->type->type_tag == NATURAL) {
			strcat(buf, "/ DROP");
		}
		break;
	case '<':
		if(op->type->type_tag == BINARY) {
			strcpy(buf, "HXS<HXS");
		} else {
			strcat(buf, "<");
		}
		break;
	case '>':
		if(op->type->type_tag == BINARY) {
			strcpy(buf, "HXS>HXS");
		} else {
			strcat(buf, ">"); 
		}
		break;
	case GE :
		if(op->type->type_tag == BINARY) {
			strcpy(buf, "HXS>=HXS");
		} else {
			strcat(buf, ">=");
		}
		break;
	case LE :
		if(op->type->type_tag == BINARY) {
			strcpy(buf, "HXS<=HXS");
		} else {
			strcat(buf, "<="); 
		}
		break;
	case NE :
		if(op->type->type_tag != BINARY) {
			strcat(buf, "<>");
		} else {
			strcpy(buf, "HXS#HXS");
		}
		break;
	case EQUAL:
		if(op->type->type_tag != BINARY) {
			strcat(buf, "=");
		} else {
			strcpy(buf, "HXS==HXS");
		}
		break;
	case UMINUS: 
		switch(op->type->type_tag) {
			case REAL:			strcpy(buf, opt_t ? "%%CHS" : "%CHS"); break;
			case LONGREAL:		strcpy(buf, opt_t ? "%%%%CHS" : "%%CHS" ); break;
			case COMPLEX:		strcpy(buf, opt_t ? "C%%CHS" : "C%CHS"); break;
			case LONGCOMPLEX:	strcpy(buf, opt_t ? "C%%%%CHS" : "C%%CHS"); break;
			case BINARY:
				if(opt_t) { nl(out); print(out, "ASSEMBLE"); }
				print(out, "CON(5) %d", 0x53EC3);
				if(opt_t) { print(out, "RPL"); nl(out); }
				return;
			default:
				error(op->line, "invalid type");
		}
		break;
	case '&' :	strcat(buf, "AND"); break;
	case '|' :	strcat(buf, "OR"); break;
	case '~' :	strcat(buf, "NOT"); break;
	case SHL :	strcat(buf, "SL"); break;
	case SHR :	strcat(buf, "SR"); break;
	case ASR :	strcat(buf, "ASR"); break;
	case ROL :	strcat(buf, "RL"); break;
	case ROR :	strcat(buf, "RR"); break;
	}
	print(out, buf);
}

/*	Determines casting operator based on the two types.  */
static void cast(struct st_node *op)
{
	int to = op->type->type_tag, from;
	
	if(op->class.control.children[0]->type == NULL) return;
	from = op->class.control.children[0]->type->type_tag;

	switch(from) {
	case NATURAL:
		if(to == REAL) print(out, "UNCOERCE");
		else if(to == STRING) print(out, "#>$");
		else if(to == BINARY) print(out, "#>HXS");
		else if(to == CHAR) print(out, "#>CHR");
		else error(op->line, "invalid cast");
		break;
	case REAL:
		if(to == NATURAL) {
			warning(op->line, "conversion from REAL to NATURAL REPLACES negative values with 0");
			print(out, "COERCE");
		}
		else if(to == COMPLEX) print(out, opt_t ? "Re>C%%" : "Re>C%");
		else if(to == STRING) print(out, opt_t ? "a%%>$" : "a%>$");
		else if(to == BINARY) {
			warning(op->line, "conversion from REAL to BINARY ROUNDS decimals and REPLACES negative values with 0");
			print(out, opt_t ? "%%>#" : "%>#");
		}
		else error(op->line, "invalid cast");
		break;
	case BINARY:
		if(to == NATURAL) print(out, "HXS>#");
		else if(to == REAL) print(out, opt_t ? "HXS>%%" : "HXS>%");
		else if(to == STRING) {
			warning(op->line, "conversion from BINARY to STRING will NOT append base character");
			print(out, "hxs>$");
		}
		else error(op->line, "invalid cast");
		break;
	case CHAR:
		if(to == NATURAL) print(out, "CHR>#");
		else if(to == STRING) print(out, "CHR>$");
		else error(op->line, "invalid cast");
		break;
	default:
		error(op->line, "invalid cast");
	}
}

/*	Code generation. */
static void compile_tree(struct st_node *op)
{
	if(!op) return;

	switch(op->opcode) {
	case ';':		/* statements */
	case ',':
		compile_tree(op->class.control.children[0]);
		nl(out);
		compile_tree(op->class.control.children[1]);
		break;
	case ASSIGN:	/* IDN, expr */
		compile_tree(op->class.control.children[1]);
		print(out, "'");
		compile_tree(op->class.control.children[0]);
		print(out, "STO");
		break;
	case ARRGET:	/* IDN, expr */
		compile_tree(op->class.control.children[1]);
		compile_tree(op->class.control.children[0]);
		print(out, "GETATELN DROP");	/* removes TRUE/FALSE flag */
		break;
	case ARRPUT:	/* IDN, expr, expr */
		compile_tree(op->class.control.children[0]);
		compile_tree(op->class.control.children[2]);
		compile_tree(op->class.control.children[1]);
		print(out, "PUTEL '");	/* ' is for the following IDN */
		compile_tree(op->class.control.children[0]);
		print(out, "STO");
		break;
	case PROCEDURE:	/* IDN, actualparm */
	case FUNCTION:
		/* calculate parameters and then the identifier */
		compile_tree(op->class.control.children[1]);
		compile_tree(op->class.control.children[0]);
		break;
	case IF:		/* boolexpr, iftrue */
		compile_tree(op->class.control.children[0]);
		print(out, "IT ::"); ++indent_level; nl(out);
		compile_tree(op->class.control.children[1]);
		--indent_level; nl(out); print(out, ";"); nl(out);
		break;
	case ELSE:		/* boolexpr, iftrue, iffalse */
		compile_tree(op->class.control.children[0]);
		print(out, "ITE"); ++indent_level; nl(out);
		print(out, "::"); ++indent_level; nl(out);
		compile_tree(op->class.control.children[1]);
		--indent_level; nl(out); print(out, ";");
		nl(out); print(out, "::");
		++indent_level; nl(out);
		compile_tree(op->class.control.children[2]);
		--indent_level; nl(out); print(out, ";");
		--indent_level; nl(out);
		break;
	case CASE:		/* condlist */
		print(out, "::"); ++indent_level; nl(out);
		compile_tree(op->class.control.children[0]);
		--indent_level; nl(out); print(out, ";"); nl(out);
		break;
	case THEN:		/* (casecond) boolexpr, statements */
		compile_tree(op->class.control.children[0]);
		print(out, "case ::"); ++indent_level; nl(out);
		compile_tree(op->class.control.children[1]);
		--indent_level; nl(out); print(out, ";"); nl(out);
		break;
	case START:		/* expr, expr, statements, expr */
		/* calculate stop value and increment it by one */
		compile_tree(op->class.control.children[1]);
		print(out, "#1+");
		compile_tree(op->class.control.children[0]);
		print(out, "DO"); ++indent_level; nl(out);
		if(lstk >= 255) FATAL("loop stack overflow");
		loopenv[++lstk] = DO;
		compile_tree(op->class.control.children[2]);
		compile_tree(op->class.control.children[3]);
		--lstk;
		--indent_level; nl(out); print(out, "+LOOP"); nl(out);
		break;
	case NEXT:		/* expr, expr, statements */
		compile_tree(op->class.control.children[1]);
		print(out, "#1+");
		compile_tree(op->class.control.children[0]);
		print(out, "DO"); ++indent_level; nl(out);
		if(lstk >= 255) FATAL("loop stack overflow");
		loopenv[++lstk] = DO;
		compile_tree(op->class.control.children[2]);
		--lstk;
		--indent_level; nl(out); print(out, "LOOP"); nl(out);
		break;
	case DO:
		print(out, "BEGIN"); ++indent_level; nl(out);
		if(lstk >= 255) FATAL("loop stack overflow");
		loopenv[++lstk] = UNTIL;
		compile_tree(op->class.control.children[0]);
		--lstk;
		--indent_level; nl(out);
		compile_tree(op->class.control.children[1]);
		print(out, "UNTIL"); nl(out);
		break;
	case WHILE:
		print(out, "BEGIN");
		compile_tree(op->class.control.children[0]);
		print(out, "WHILE ::"); ++indent_level; nl(out);
		if(lstk >= 255) FATAL("loop stack overflow");
		loopenv[++lstk] = WHILE;
		compile_tree(op->class.control.children[1]);
		--lstk;
		--indent_level; nl(out); print(out, "; REPEAT"); nl(out);
		break;
	case AND:
		compile_tree(op->class.control.children[0]);
		compile_tree(op->class.control.children[1]);
		print(out, "AND");
		break;
	case ANDTHEN:	/* short-circuit AND */
		compile_tree(op->class.control.children[0]);
		print(out, "DUP IT ::");
		compile_tree(op->class.control.children[1]);
		print(out, "AND ;");
		break;
	case OR:
		compile_tree(op->class.control.children[0]);
		compile_tree(op->class.control.children[1]);
		print(out, "OR");
		break;
	case ORELSE:	/* short-circuit OR */
		compile_tree(op->class.control.children[0]);
		print(out, "DUP ?SKIP ::");
		compile_tree(op->class.control.children[1]);
		print(out, "OR ;");
		break;
	case NOT:
		compile_tree(op->class.control.children[0]);
		print(out, "NOT");
		break;
	case '>': case '<': case GE: case LE: case NE: case EQUAL:
	case '+': case '-': case '*': case '/': case '&': case '|':
		compile_tree(op->class.control.children[0]);
		compile_tree(op->class.control.children[1]);
		oper(op);
		break;
	case UMINUS:
	case '~': case SHL: case SHR: case ASR: case ROL: case ROR:
		compile_tree(op->class.control.children[0]);
		oper(op);
		break;
	case LIT_NATURAL:
		print(out, "# %x", op->class.literal.natural);
		break;
	case LIT_BOOLEAN:
		print(out, op->class.literal.boolean ? "TRUE" : "FALSE");
		break;
	case LIT_REAL:
		print(out, "%% %f", op->class.literal.real);
		break;
	case LIT_COMPLEX:
		print(out, "C%% %f %f", op->class.literal.complex[0], op->class.literal.complex[1]);
		break;
	case LIT_BINARY:
		print(out, "HXS %x %s", strlen(op->class.literal.binary), op->class.literal.binary);
		break;
	case LIT_CHAR:
		print(out, "CHR \\%02x", op->class.literal.chr);
		break;
	case LIT_STRING:
		print(out, "$ \"%s\"", op->class.literal.str);
		break;
	case IDN:
		if(op->type->type_tag != PROCEDURE && op->type->type_tag != FUNCTION) {
			if(op->type->scope == GLOBAL) {
				print(out, "ID %s", op->class.identifier.name);
			} else {
				print(out, "LAM %s", op->class.identifier.name);
			}
		} else {
			if(op->type->class.subroutine.external >= 0) {
				if(opt_t) { nl(out); print(out, "ASSEMBLE"); }
				print(out, "CON(5) %d", op->type->class.subroutine.external);
				if(opt_t) { print(out, "RPL * %s", op->class.identifier.name); nl(out); }
			} else {
				print(out, "ID %s", op->class.identifier.name);
			}
			nl(out);
		}
		break;
	case OVER:	print(out, "OVER"); break;
	case DUP:	print(out, "DUP"); break;
	case SWAP:	print(out, "SWAP"); break;
	case ROT:	print(out, "ROT"); break;
	case DROP:	print(out, "DROP"); break;
	case DEPTH:	print(out, "DEPTH"); break;
	case PUSH:
		compile_tree(op->class.control.children[0]);
		break;
	case ROLL:
		compile_tree(op->class.control.children[0]);
		print(out, "ROLL");
		break;
	case ROLLD:
		compile_tree(op->class.control.children[0]);
		print(out, "UNROLL");
		break;
	case ':':
		compile_tree(op->class.control.children[0]);
		print(out, "PICK");
		break;
	case INDEX0:
		if(lstk <= 0 || loopenv[lstk] != DO) error(op->line, "no matching START for `@0'");
		print(out, "INDEX@");
		break;
	case INDEX1:
		if(lstk <=1 || loopenv[lstk-1] != DO) error(op->line, "no matching START for `@1'");
		print(out, "JINDEX@");
		break;
	case NATURAL:
	case REAL:
	case STRING:
	case BINARY:
	case CHAR:
	case COMPLEX:
		compile_tree(op->class.control.children[0]);
		cast(op);
		break;
	case RE2COMPLEX:
		compile_tree(op->class.control.children[0]);
		compile_tree(op->class.control.children[1]);
		print(out, opt_t ? "%%>C%%" : "%>C%");
		break;
	case TOP:
		break;
	case RE:
		compile_tree(op->class.control.children[0]);
		print(out, opt_t ? "C>Re%%" : "C>Re%");
		break;
	case IM:
		compile_tree(op->class.control.children[0]);
		print(out, opt_t ? "C>Im%%" : "C>Im%");
		break;
	case BREAK:
		if(lstk <= 0) error(op->line, "break allowed only inside loops");
		if(loopenv[lstk] == DO) {
			warning(op->line, "`break' inside START loop is effective only at end of loop (NEXT/STEP)");
			print(out, "ExitAtLOOP");
		} else if(loopenv[lstk] == UNTIL) {
			error(op->line, "unimplemented `break' inside DO/UNTIL");
		} else if(loopenv[lstk] == WHILE) {
			error(op->line, "unimplemented `break' inside WHILE/REPEAT");
		} else {
			FATAL("invalid loop on top of loop stack");
		}
		break;
	case TYPEOF:	/* stackop, type */
		compile_tree(op->class.control.children[0]);
		switch(op->class.control.children[1]->type->type_tag) {
		case REAL:		print(out, "DTYPEREAL?"); break;
		case NATURAL:	print(out, "DTYPEBINT?"); break;
		case STRING:	print(out, "DTYPECSTR?"); break;
		case BINARY:	print(out, "DTYPEHSTR?"); break;
		case CHAR:		print(out, "DTYPECHAR?"); break;
		case COMPLEX:	print(out, "DTYPECMP?"); break;
		case ARRAY:
			if(op->class.control.children[1]->type->class.array.elemtype->type_tag == REAL) {
				print(out, "DTYPEARRY?");
			} else if(op->class.control.children[1]->type->class.array.elemtype->type_tag == COMPLEX) {
				print(out, "DTYPECARRY?");
			} else {
				FATAL("invalid array type");
			}
			break;
		case LIST:
			print(out, "DTYPELIST?"); break;
		}
		break;
		case KW_NULL:
			if(op->type->type_tag == LIST) {
				print(out, "NULL{}");
			} else if(op->type->type_tag == STRING) {
				print(out, "NULL$");
			} else {
				FATAL("invalid NULL type");
			}
			break;
		case PROGRAM:
			/*	So that we don't have to parse complicated formats when
			 *	writing executable code in saturn_print(). */
			sprintf(buf, "CK%d&Dispatch", (int)op->class.control.children[1]);
			print(out, buf);
			compile_tree(op->class.control.children[0]);
			break;
		case DISPATCH:
			print(out, "# %x", (int)op->class.control.children[0]);
			print(out, "ID %s", op->class.control.children[1]->class.identifier.name);
			break;
	default:
		FATAL("unimplemented operation");
		exit(1);
	}
}

