/*
 * hpc.l - LEX lexical scanner
 * (c) 2000 Zeljko Vrba (mordor@fly.srk.fer.hr)
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

%option case-insensitive yylineno

%{
#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include "node.h"
#include "util.h"
#include "hpc-y.h"

#ifdef __GNUC__
static char rcsid[] __attribute__((unused)) = "$Id: hpc-l.l,v 1.1 2000-01-08 09:23:06+01 zeljko Exp $";
#else
static char rcsid[] = "$Id: hpc-l.l,v 1.1 2000-01-08 09:23:06+01 zeljko Exp $";
#endif
%}

%%

	/* comments */
--[^\n]*			;

	/* whitespace */
[ \t\n]				;

	/* natural numbers */
[0-9]+				{
						yylval.op = new_st_literal(LIT_NATURAL, atoi(yytext));
						if(yylval.op->class.literal.natural > 0xFFFFF) {
							error(yylineno, "natural constant out of range");
						}
						return LIT_NATURAL;
					}

	/* natural numbers in hexadecimal base */
0x[a-f0-9]+			{
						yylval.op = new_st_literal(LIT_NATURAL, strtol(yytext, NULL, 16));
						if(yylval.op->class.literal.natural > 0xFFFFF) {
							error(yylineno, "natural constant out of range");
						}
						return LIT_NATURAL;
					}

	/* real numbers */
[0-9]+"."([0-9]*)?([e|E]-?[0-9]+)? {
						yylval.op = new_st_literal(LIT_REAL, atof(yytext));
						return LIT_REAL;
					}
					
	/* binary number (hex string) */
"#"[a-f0-9]+		{
						/* without leading # */
						yylval.op = new_st_literal(LIT_BINARY, yytext+1);
						return LIT_BINARY;
					}

	/* character constant */
"'"."'"				{
						yylval.op = new_st_literal(LIT_CHAR, yytext[1]);
						return LIT_CHAR;
					}

	/* string constant */
\"[^\"\n]*\"		{
						/* remove trailing " delimiter */
						yytext[yyleng-1] = 0;
						yylval.op = new_st_literal(LIT_STRING, yytext+1);
						return LIT_STRING;
					}

	/* boolean constant */
false				{
						yylval.op = new_st_literal(LIT_BOOLEAN, 0);
						return LIT_BOOLEAN;
					}
true				{
						yylval.op = new_st_literal(LIT_BOOLEAN, 1);
						return LIT_BOOLEAN;
					}

	/* null list or null string? depends on context; the parser decides */
null				return KW_NULL;

	/* keywords for flow control */
if					return IF;
then				return THEN; 
else				return ELSE;
endif				return ENDIF;
case				return CASE; 
do					return DO;
until				return UNTIL;
while				return WHILE;
repeat				return REPEAT;
start				return START;
step				return STEP;
next				return NEXT;
begin				return KWBEGIN;
end					return END;
break				{ yylval.op = new_st_control(BREAK, NULL, NULL); return BREAK; }

	/* declaration keywords */
program				return PROGRAM;
var					return VAR;
unbound				return UNBOUND;
external			return EXTERNAL;
at					return AT;
function			return FUNCTION;
procedure			return PROCEDURE;
typeof				return TYPEOF;
array				return ARRAY;
of					return OF;
list				return LIST;

	/* type keywords */
real				{ yylval.ty = new_type(REAL); return REAL; }
longreal			{ yylval.ty = new_type(LONGREAL); return LONGREAL; }
natural				{ yylval.ty = new_type(NATURAL); return NATURAL; }
string				{ yylval.ty = new_type(STRING); return STRING; }
binary				{ yylval.ty = new_type(BINARY); return BINARY; }
char				{ yylval.ty = new_type(CHAR); return CHAR; }
complex				{ yylval.ty = new_type(COMPLEX); return COMPLEX; }
longcomplex			{ yylval.ty = new_type(LONGCOMPLEX); return LONGCOMPLEX; }

	/* operators */
"+"|"-"|"*"|"/"|"[" |
"]"|"."|"<"|">"|":" |
";"|","|"("|")"|"&"	|
"|"|"~"|"~"|"%"		|
"`"					return *yytext;
"@0"				{ yylval.op = new_st_control(INDEX0, new_type(NATURAL), NULL); return INDEX0; }
"@1"				{ yylval.op = new_st_control(INDEX1, new_type(NATURAL), NULL); return INDEX1; }
shl					return SHL;
shr					return SHR;
asr					return ASR;
rol					return ROL;
ror					return ROR;
and					return AND;
or					return OR;
not					return NOT;
">="				return GE;
"<="				return LE;
"!="				return NE;
"="					return ASSIGN;
"=="				return EQUAL;
re					return RE;
im					return IM;

	/* stack operations */
over				{ yylval.op = new_st_control(OVER, NULL, NULL); return OVER; }
dup					{ yylval.op = new_st_control(DUP, NULL, NULL); return DUP; }
drop				{ yylval.op = new_st_control(DROP, NULL, NULL); return DROP; }
swap				{ yylval.op = new_st_control(SWAP, NULL, NULL); return SWAP; }
rot					{ yylval.op = new_st_control(ROT, NULL, NULL); return ROT; }
depth				{ yylval.op = new_st_control(DEPTH, NULL, NULL); return DEPTH; }
"()"				return TOP;
roll				return ROLL;
rolld				return ROLLD;

	/* this must be at the end; otherwise a keyword couldn't be recognized */
[\$a-z][\$a-z0-9_]*		{
						int i;

						/* convert to uppercase */
						for(i = 0; i < yyleng; i++) yytext[i] = toupper(yytext[i]);
						yylval.op = new_st_identifier(yytext);

						/* find the type in symbol tables */
						yylval.op->type = find_name(yytext);

						return IDN;
					}

.					{ error(yylineno, "invalid character"); }

%%
int yywrap()
{
	return 1;
}
