/*
 * hpc.y - YACC parser source
 * (c) 2000 Zeljko Vrba (mordor@fly.srk.fer.hr)
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

%{
#include <config.h>
#include <stdarg.h>
#include "node.h"
#include "util.h"

extern int yylineno;
extern int yylex();
extern void output_code(const char*, struct st_node*);
static int type_nybble(struct type*);
static int type_check(struct type*, struct type*, int);
static void arg_type_check(const char*);
static int code_dispatch_list(void);
static void yyerror(const char*);

/*	Checks whether a type has an ordering relation, like < */
static inline int is_ordered(int tag) {
	return tag == NATURAL
		|| tag == REAL
		|| tag == LONGREAL
		|| tag == BINARY;
}

#ifdef __GNUC__
static char rcsid[] __attribute__((unused)) = "$Id: hpc-y.y,v 1.1 2000-01-08 09:22:59+01 zeljko Exp $";
#else
static char rcsid[] = "$Id: hpc-y.y,v 1.1 2000-01-08 09:22:59+01 zeljko Exp $";
#endif
%}

%union {
	struct st_node *op;
	struct type    *ty;
}

%{
/* stack that holds parameter or argument lists */
static struct st_node *argstk[256];
static int sp = -1;

/* current declaration scope */
static enum scope scope = GLOBAL;
%}

%token			PROGRAM FUNCTION PROCEDURE VAR UNBOUND EXTERNAL 
%token			AT OF TYPEOF IF THEN ENDIF ELSE CASE START DO STEP NEXT
%token			UNTIL WHILE	REPEAT KWBEGIN END TOP KW_NULL

%token	<ty>	REAL NATURAL CHAR STRING BINARY
%token	<ty>	LIST COMPLEX ARRAY LONGREAL LONGCOMPLEX
%token	<op>	RE IM INDEX0 INDEX1 BREAK
%token	<op>	OVER DUP SWAP ROT ROLL ROLLD DEPTH DROP
%token	<op>	LIT_NATURAL LIT_REAL LIT_BINARY LIT_BOOLEAN
%token	<op>	LIT_CHAR LIT_STRING IDN

%left	','
%right	ASSIGN
%left	OR
%left	AND
%left	NOT
%left	'<' '>' EQUAL GE LE NE
%left	SHL SHR ASR ROR ROL '&' '|' '~'
%left	'+' '-'
%left	'*' '/' '%'
%nonassoc UMINUS RE IM

%type	<op>	statement statements stackop expr boolexpr arithexp 
%type	<op>	condlist casecond  actualparm dispatch exprlist
%type	<op>	function funchdr procedure prochdr proghdr
%type	<ty>	type simpletype

%%

unit		:	vardecl subroutines
			;

vardecl		:	vardecl1
			|	/* empty */
			;
			
vardecl1	:	varstmt
			|	vardecl1 varstmt
			;

varstmt		:	VAR typedfn ';'
			;

typedfn		:	varlist ':' type
				{
					int i;
					
					/* All variables in varlist are on the stack. Assign them
					 * type and put them in the symbol table. */
					for(i = 0; i <= sp; i++) {
						argstk[i]->type = $3;
						argstk[i]->type->scope = scope;
						if(scope == GLOBAL) {
							if(!add_global_name(argstk[i]->class.identifier.name, argstk[i]->type)) error(yylineno, "duplicate identifier");
							argstk[i]->type->pos = -1;
						} else {
							/* parameters and local variables go together */
							argstk[i]->type->pos = i;
							if(!add_local_name(argstk[i]->class.identifier.name, argstk[i]->type)) error(yylineno, "duplicate identifier");
						}
					}
					sp = -1;
				}
			;

varlist		:	IDN
				{
					if(sp >= 256) error(yylineno, "argument stack xceeded");
					argstk[++sp] = $1;
				}
			|	varlist ',' IDN
				{
					if(sp >= 256) error(yylineno, "argument stack exceeded");
					argstk[++sp] = $3;
				}
			;

type		:	simpletype
			|	ARRAY OF simpletype
				{
					if($3->type_tag != REAL && $3->type_tag !=COMPLEX) error(yylineno, "only REAL or COMPLEX arrays allowed");
					$$ = new_type(ARRAY, $3, 0);
				}
			|	ARRAY '[' LIT_NATURAL ']' OF simpletype
				{
					if($6->type_tag != REAL && $6->type_tag !=COMPLEX) error(yylineno, "only REAL or COMPLEX arrays allowed");
					$$ = new_type(ARRAY, $6, 1, $3->class.literal.natural);
				}
			|	ARRAY '[' LIT_NATURAL ',' LIT_NATURAL ']' OF simpletype
				{
					if($8->type_tag != REAL && $8->type_tag !=COMPLEX) error(yylineno, "only REAL or COMPLEX arrays allowed");
					$$ = new_type(ARRAY, $8, 2, $3->class.literal.natural, $5->class.literal.natural);
				}
			|	LIST
				{
					$$ = new_type(LIST);
				}
			;

simpletype	:	REAL
			|	LONGREAL
			|	NATURAL
			|	STRING
			|	BINARY
			|	CHAR
			|	COMPLEX
			|	LONGCOMPLEX
			;

subroutines	:	subrlist
			|	/* empty */
			;

subrlist	:	function	{ }
			|	procedure	{ }
			|	program { }
			|	subrlist function
			|	subrlist procedure
			|	subrlist program
			;

function	:	EXTERNAL { scope = PARAM; } funchdr AT LIT_NATURAL ';'
				{
					if($5->class.literal.natural < 0 || $5->class.literal.natural > 0xFFFFF) error(yylineno, "invalid address");
					$3->type->class.subroutine.external = $5->class.literal.natural;
					clear_local_names();
				}
			|	{ scope = PARAM; } funchdr ';' { scope = LOCAL; } vardecl KWBEGIN statements END ';'
				{
					output_code($2->class.identifier.name, $7);
					clear_local_names();
				}
			;

funchdr		:	FUNCTION IDN '(' formalparm ')' ':' type
				{
					/* there are no local variables other than parameters. */
					$2->type = new_type(FUNCTION, $7);
					if(!add_global_name($2->class.identifier.name, $2->type)) error(yylineno, "duplicate identifier");
					$$ = $2;
				}
			|	FUNCTION IDN ':' type 
				{
					$2->type = new_type(FUNCTION, $4);
					if(!add_global_name($2->class.identifier.name, $2->type)) error(yylineno, "duplicate identifier");
					$$ = $2;
				}
			;
			

procedure	: 	EXTERNAL { scope = PARAM; } prochdr AT LIT_NATURAL ';'
				{
					if($5->class.literal.natural < 0 || $5->class.literal.natural > 0xFFFFF) error(yylineno, "invalid address");
					$3->type->class.subroutine.external = $5->class.literal.natural;
					clear_local_names();
				}
			|	{ scope = PARAM; } prochdr ';' { scope = LOCAL; } vardecl KWBEGIN statements END ';'
				{
					output_code($2->class.identifier.name, $7);
					clear_local_names();
				}
			;
			
prochdr		:	PROCEDURE IDN '(' formalparm ')'
				{
					/* there are no local variables other than parameters. */
					$2->type = new_type(PROCEDURE);
					if(!add_global_name($2->class.identifier.name, $2->type)) error(yylineno, "duplicate identifier");
					$$ = $2;
				}
			|	PROCEDURE IDN 
				{
					$2->type = new_type(PROCEDURE);
					if(!add_global_name($2->class.identifier.name, $2->type)) error(yylineno, "duplicate identifier");
					$$ = $2;
				}
			;

program		:	proghdr KWBEGIN dispatch END ';'
				{ output_code($1->class.identifier.name, new_st_control(PROGRAM, NULL, $3, $1->type->class.subroutine.nargs, NULL)); }

proghdr		:	PROGRAM IDN '(' LIT_NATURAL ')' ';'
				{
					$2->type = new_type(PROGRAM, $4->class.literal.natural);
					if($4->class.literal.natural < 0 || $4->class.literal.natural > 5) error(yylineno, "invalid number for dispatch list");
					if(!add_global_name($2->class.identifier.name, $2->type)) error(yylineno, "duplicate identifier"); 
					$$ = $2;
				}

dispatch	:	typelist ':' IDN 
				{
					int type_code;
					
					if($3->type == NULL) error(yylineno, "undeclared identifier");
					if($3->type->type_tag != PROCEDURE && $3->type->type_tag != FUNCTION) error(yylineno, "not a procedure/function");
					if(sp > 4) error(yylineno, "programs can be dispatched on at most 5 arguments");
					arg_type_check($3->class.identifier.name);
					type_code = code_dispatch_list();
					$$ = new_st_control(DISPATCH, NULL, type_code, $3, NULL);
					sp = -1;
				}
			|	dispatch typelist ':' IDN 
				{
					int type_code;
					
					if($4->type == NULL) error(yylineno, "undeclared identifier");
					if($4->type->type_tag != PROCEDURE && $4->type->type_tag != FUNCTION) error(yylineno, "not a procedure/function");
					if(sp > 4) error(yylineno, "programs can be dispatched on at most 5 arguments");
					arg_type_check($4->class.identifier.name);
					type_code = code_dispatch_list();
					$$ = new_st_control(',', NULL, $1, new_st_control(DISPATCH, NULL, type_code, $4, NULL));
					sp = -1;
				}
			;

typelist	:	type
				{
					if(sp >= 256) error(yylineno, "type stack exceeded");
					argstk[++sp] = new_st_control(-1, $1, NULL);
				}
			|	typelist ',' type
				{
					if(sp >= 256) error(yylineno, "type stack exceeded");
					argstk[++sp] = new_st_control(-1, $3, NULL);
				}
			;

formalparm	:	typedfn
			|	formalparm ';' typedfn
			;

statements	:	statement ';'
			|	statements statement ';'
				{ $$ = new_st_control(';', NULL, $1, $2, NULL); }
			;

statement	:	IDN ASSIGN expr 
				{
					/* lexical analyzer locates the IDN in the symbol table */
					if($1->type == NULL) error(yylineno, "undeclared identifier");
					type_check($1->type, $3->type, 1);
					$$ = new_st_control(ASSIGN, NULL, $1, $3, NULL);
				}
			|	IDN '[' expr ']' ASSIGN expr
				{
					if($1->type == NULL) error(yylineno, "undeclared identifier");
					if($1->type->type_tag != ARRAY) error(yylineno, "not an array");
					if($3->type->type_tag != NATURAL) error(yylineno, "array index is not an natural");
					type_check($1->type->class.array.elemtype, $6->type, 1);
					$$ = new_st_control(ARRPUT, NULL, $1, $3, $6, NULL);
				}
			|	IDN actualparm
				{
					if($1->type == NULL) error(yylineno, "undeclared identifier");
					if($1->type->type_tag != PROCEDURE) error(yylineno, "not a procedure call");
					arg_type_check($1->class.identifier.name);
					$$ = new_st_control(PROCEDURE, NULL, $1, $2, NULL);
					sp = -1;
				}
			|	IF boolexpr THEN statements ENDIF
				{ $$ = new_st_control(IF, NULL, $2, $4, NULL); }
			|	IF boolexpr THEN statements ELSE statements ENDIF
				{ $$ = new_st_control(ELSE, NULL, $2, $4, $6, NULL); }
			|	CASE condlist END
				{ $$ = new_st_control(CASE, NULL, $2, NULL); }
			|	START expr ',' expr DO statements STEP expr
				{
					if(($2->type->type_tag != NATURAL)
					|| ($4->type->type_tag != NATURAL)
					|| ($8->type->type_tag != NATURAL)) {
						error(yylineno, "natural expression required for loop bounds/step expression");
					}
					$$ = new_st_control(START, NULL, $2, $4, $6, $8, NULL);
				}
			|	START expr ',' expr DO statements NEXT
				{
					if(($2->type->type_tag != NATURAL)
					|| ($4->type->type_tag != NATURAL)) {
						error(yylineno, "natural expression required for loop bounds");
					}
					$$ = new_st_control(NEXT, NULL, $2, $4, $6, NULL);
				}
			|	DO statements UNTIL boolexpr
				{ $$ = new_st_control(DO, NULL, $2, $4, NULL); }
			|	WHILE boolexpr REPEAT statements END
				{ $$ = new_st_control(WHILE, NULL, $2, $4, NULL); }
			|	TOP ASSIGN expr
				{ $$ = new_st_control(PUSH, NULL, $3, NULL); }
			|	BREAK
			|	stackop
			;

condlist	:	casecond
				{ $$ = new_st_control(',', NULL, $1, NULL); }
			|	condlist casecond
				{ $$ = new_st_control(',', NULL, $1, $2, NULL); }
			;

casecond	:	boolexpr THEN statements END ';'
				{ $$ = new_st_control(THEN, NULL, $1, $3, NULL); }
			;

stackop		:	OVER
			|	DUP
			|	SWAP
			|	ROT
			|	DEPTH
			|	DROP
			|	ROLL ':' expr
				{
					if($3->type->type_tag != NATURAL) error(yylineno, "natural expression required for stack depth");
					$$ = new_st_control(ROLL, NULL, $3, NULL);
				}
			|	ROLLD ':' expr
				{
					if($3->type->type_tag != NATURAL) error(yylineno, "natural expression required for stack depth");
					$$ = new_st_control(ROLLD, NULL, $3, NULL);
				}
			|	':' expr
				{
					if($2->type->type_tag != NATURAL) error(yylineno, "natural expression required for stack depth");
					$$ = new_st_control(':', NULL, $2, NULL);
				}
			;
			
boolexpr	:	LIT_BOOLEAN
			|	boolexpr AND boolexpr
				{ $$ = new_st_control(AND, NULL, $1, $3, NULL); }
			|	boolexpr OR boolexpr
				{ $$ = new_st_control(OR, NULL, $1, $3, NULL); }
			|	NOT boolexpr
				{ $$ = new_st_control(NOT, NULL, $2, NULL); }
			|	'(' boolexpr ')'
				{ $$ = $2; }
			|	boolexpr AND THEN boolexpr %prec AND
				{ $$ = new_st_control(ANDTHEN, NULL, $1, $4, NULL); }
			|	boolexpr OR ELSE boolexpr %prec OR
				{ $$ = new_st_control(ORELSE, NULL, $1, $4, NULL); }
			|	expr '>' expr
				{
					type_check($1->type, $3->type, 1);
					if(!is_ordered($1->type->type_tag)) error(yylineno, "invalid type");
					$$ = new_st_control('>', $1->type, $1, $3, NULL);
				}
			|	expr '<' expr
				{
					type_check($1->type, $3->type, 1);
					if(!is_ordered($1->type->type_tag)) error(yylineno, "invalid type");
					$$ = new_st_control('<', $1->type, $1, $3, NULL);
				}
			|	expr GE expr
				{
					type_check($1->type, $3->type, 1);
					if(!is_ordered($1->type->type_tag)) error(yylineno, "invalid type");
					$$ = new_st_control(GE, $1->type, $1, $3, NULL);
				}
			|	expr LE expr
				{
					type_check($1->type, $3->type, 1);
					if(!is_ordered($1->type->type_tag)) error(yylineno, "invalid type");
					$$ = new_st_control(LE, $1->type, $1, $3, NULL);
				}
			|	expr NE expr
				{
					type_check($1->type, $3->type, 1);
					$$ = new_st_control(NE, $1->type, $1, $3, NULL);
				}
			|	expr EQUAL expr
				{
					type_check($1->type, $3->type, 1);
					$$ = new_st_control(EQUAL, $1->type, $1, $3, NULL);
				}
			|	TYPEOF stackop TOP EQUAL type
				{ $$ = new_st_control(TYPEOF, NULL, $2, $5, NULL); }
			;

expr		:	arithexp
			|	INDEX0
			|	INDEX1
			|	type TOP
				{ $$ = new_st_control(TOP, $1, NULL); }
			|	type stackop TOP
				{ $$ = $2; $$->type = $1; }
			|	type expr %prec UMINUS	/* casting */
				{
					if($2->type != NULL) {
						$$ = new_st_control($1->type_tag, $1, $2, NULL);
					} else {
						$$ = $2;
					}
				}
			|	COMPLEX '(' expr ',' expr ')' 
				{
					if($3->type->type_tag != REAL || $5->type->type_tag != REAL) error(yylineno, "two reals expected in `complex'");
					$$ = new_st_control(RE2COMPLEX, new_type(COMPLEX), $3, $5, NULL);
				}
			|	RE expr
				{
					if($2->type->type_tag != COMPLEX) error(yylineno, "`re' can only be used on COMPLEX expressions");
					$$ = new_st_control(RE, new_type(REAL), $2, NULL);
				}
			|	IM expr
				{
					if($2->type->type_tag != COMPLEX) error(yylineno, "`im' can be used only on COMPLEX expressions");
					$$ = new_st_control(IM, new_type(REAL), $2, NULL);
				}
			|	KW_NULL LIST
				{ $$ = new_st_control(KW_NULL, new_type(LIST), NULL); }
			|	KW_NULL STRING
				{ $$ = new_st_control(KW_NULL, new_type(STRING), NULL); }
			|	LIT_STRING
			|	LIT_CHAR
			;
			
arithexp	:	LIT_NATURAL
			|	LIT_REAL
			|	LIT_BINARY
			|	IDN
			|	expr '+' expr
				{
					type_check($1->type, $3->type, 1);
					$$ = new_st_control('+', $1->type, $1, $3, NULL);
				}
			|	expr '-' expr
				{
					type_check($1->type, $3->type, 1);
					$$ = new_st_control('-', $1->type, $1, $3, NULL);
				}
			|	expr '*' expr
				{
					type_check($1->type, $3->type, 1);
					$$ = new_st_control('*', $1->type, $1, $3, NULL);
				}
			|	expr '/' expr
				{
					type_check($1->type, $3->type, 1);
					$$ = new_st_control('/', $1->type, $1, $3, NULL);
				}
			|	expr '%' expr
				{
					type_check($1->type, $3->type, 1);
					if($1->type->type_tag != NATURAL && $1->type->type_tag != REAL) error(yylineno, "invalid type");
					$$ = new_st_control('%', $1->type, $1, $3, NULL);
				}
			|	expr '&' expr
				{
					type_check($1->type, $3->type, 1);
					if($1->type->type_tag != BINARY) error(yylineno, "invalid type");
					$$ = new_st_control('&', $1->type, $1, $3, NULL);
				}
			|	expr '|' expr
				{
					type_check($1->type, $3->type, 1);
					if($1->type->type_tag != BINARY) error(yylineno, "invalid type");
					$$ = new_st_control('|', $1->type, $1, $3, NULL);
				}
			|	'~' expr
				{
					if($2->type->type_tag != BINARY) error(yylineno, "invalid type");
					$$ = new_st_control('~', $2->type, $2, NULL);
				}
			|	SHL expr
				{
					if($2->type->type_tag != BINARY) error(yylineno, "invalid type");
					$$ = new_st_control(SHL, $2->type, $2, NULL);
				}
			|	SHR expr
				{
					if($2->type->type_tag != BINARY) error(yylineno, "invalid type");
					$$ = new_st_control(SHR, $2->type, $2, NULL);
				}
			|	ASR expr
				{
					if($2->type->type_tag != BINARY) error(yylineno, "invalid type");
					$$ = new_st_control(ASR, $2->type, $2, NULL);
				}
			|	ROL expr
				{
					if($2->type->type_tag != BINARY) error(yylineno, "invalid type");
					$$ = new_st_control(ROL, $2->type, $2, NULL);
				}
			|	ROR expr
				{
					if($2->type->type_tag != BINARY) error(yylineno, "invalid type");
					$$ = new_st_control(ROR, $2->type, $2, NULL);
				}
			|	'(' expr ')'
				{ $$ = $2; }
			|	'-' expr %prec UMINUS
				{ $$ = new_st_control(UMINUS, $2->type, $2, NULL); }
			|	IDN actualparm
				{
					if($1->type == NULL) error(yylineno, "undeclared identifier");
					if($1->type->type_tag != FUNCTION) error(yylineno, "not a function call");
					arg_type_check($1->class.identifier.name);
					$$ = new_st_control(FUNCTION, $1->type->class.subroutine.retarg, $1, $2, NULL);
					sp = -1;
				}
			|	IDN '[' expr ']'
				{
					if($1->type == NULL) error(yylineno, "undeclared identifier");
					if($1->type->type_tag != ARRAY) error(yylineno, "not an array");
					if($3->type->type_tag != NATURAL) error(yylineno, "index is not a NATURAL");
					$$ = new_st_control(ARRGET, $1->type->class.array.elemtype, $1, $3, NULL);
				}
			;

actualparm	:	'(' exprlist ')'
				{ $$ = $2; }
			|	TOP	/* this is () */
				{ $$ = NULL; }
			;

exprlist	:	expr
				{
					if(sp >= 256) error(yylineno, "argument stack exceeded");
					argstk[++sp] = $1;
					$$ = $1;
				}
			|	exprlist ',' expr
				{
					if(sp >= 256) error(yylineno, "argument stack exceeded");
					argstk[++sp] = $3;
					$$ = new_st_control(',', NULL, $1, $3, NULL);
				}
%%

/*	Check the equality of types. */
static int
type_check(struct type *t1, struct type *t2, int terminate)
{
	/* NULL pointer is no type and is compatible with every other type. */
	if(t1 == NULL || t2 == NULL) {
		return 1;
	}
	if(t1->type_tag == ARRAY) {
		if(t2->type_tag != ARRAY) return 0;
		return t1->class.array.elemtype->type_tag == t2->class.array.elemtype->type_tag;
	}
	if(!terminate) return t1->type_tag == t2->type_tag;
	if(t1->type_tag != t2->type_tag) error(yylineno, "type mismatch");
	return 1;
}

/*	conversion of  USER type into RPL nybble */
static int
type_nybble(struct type *t)
{
	switch(t->type_tag) {
		case REAL: 		return 1;
		case COMPLEX:	return 2;
		case STRING:	return 3;
		case ARRAY:		return 4;	/* SysRPL knows only about ARRAYs */
		case LIST:		return 5;
		case BINARY:	return 11;
		default:		return -1;
	}
}

static void
arg_type_check(const char *subname)
{
	int i;
	struct type *t = find_name(subname);

	if(sp+1 != t->class.subroutine.nargs) error(yylineno, "wrong number of arguments to subroutine call");
	for(i = 0; i < t->class.subroutine.nargs; i++) {
		if(!type_check(argstk[i]->type, t->class.subroutine.args[i], 0)) error(yylineno, "wrong argument type");
	}
}

static int
code_dispatch_list(void)
{
	int i, tq, type_code = 0;

	for(i = 0; i <= sp; i++) {
		if((tq = type_nybble(argstk[i]->type)) < 0) error(yylineno, "programs can be dispatched on user type only");
		type_code = (type_code << 4) | tq;
	}
	return type_code;
}

static void yyerror(const char *s)
{
	fprintf(stderr, "%d: %s\n", yylineno, s);
}

