/*
 * node.c - symbol table and memory management, syntax node creation
 * (c) 2000 Zeljko Vrba (mordor@fly.srk.fer.hr)
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <config.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stdarg.h>
#include "node.h"
#include "hpc-y.h"

#ifdef __GNUC__
static char rcsid[] __attribute__((unused)) = "$Id: node.c,v 1.1 2000-01-08 09:24:43+01 zeljko Exp $";
#else
static char rcsid[] = "$Id: node.c,v 1.1 2000-01-08 09:24:43+01 zeljko Exp $";
#endif

#ifndef HAVE_STRDUP
char *strdup(char*);
#endif

/*	Symbol table data. I don't expect a large number of identifiers, so the
 *	O(n) complexity of searching shouldn't be a problem.  */
int globalsz = 0, localsz = 0;
char **global_names = NULL, **local_names = NULL;
struct type **global_data = NULL, **local_data = NULL;

/*	Size of currently allocated data for global and local vectors. */
static int gsize = 1, lsize = 1;

/*	Error-reporting malloc() */
static void *
xmalloc(size_t size)
{
	void *ptr = malloc(size);
	if(!ptr) {
		fprintf(stderr, "virtual memory exhausted\n");
		exit(1);
	}
	return ptr;
}

/*	Reallocating the table storage, by factor of 2. which indicates the table
 *	to enlarge (1:global, 2:local). */
static void
enlarge_table(int which)
{
	if(which == 1) {
		gsize *= 2;
		global_names = realloc(global_names, gsize*sizeof(char*));
		global_data = realloc(global_data, gsize*sizeof(struct type*));
		if((global_names == NULL) || (global_data == NULL)) {
			fprintf(stderr, "virtual memory exhausted\n");
			exit(1);
		}
	} else if(which == 2) {
		lsize *= 2;
		local_names = realloc(local_names, lsize*sizeof(char*));
		local_data = realloc(local_data, lsize*sizeof(struct type*));
		if((local_names == NULL) || (local_data == NULL)) {
			fprintf(stderr, "virtual memory exhausted\n");
			exit(1);
		}
	} else {
		FATAL("invalid table specification");
	}
}

/*	Checks for an existence of symbol in a specified symbol table. Returns
 *	an index in the table or -1 if the name doesn't exist. which specifies
 *	the table to search. */
static int
find_name_index(int which, const char *name)
{
	int i, lim;
	char **names;
	
	if(which == 1) {
		names = global_names; lim = globalsz;
	} else if(which == 2) {
		names = local_names; lim = localsz;
	} else {
		FATAL("invalid table specification");
	}
	for(i = 0; i < lim; i++) {
		if(strcmp(name, names[i]) == 0) return i;
	}
	return -1;
}


/*	Find a symbol in a symbol table. First the local table is searched, then
 *	global. NULL is returned if the name does not exist. */
struct type *
find_name(const char *name)
{
	int idx;

	if((idx = find_name_index(2, name)) >= 0) {
		return local_data[idx];
	}
	return (idx = find_name_index(1, name)) < 0 ? NULL : global_data[idx];
}

/*	Add global/local variable. data is the data associated with the declared
 *	variable. Returns 0 if the symbol wasn't added (already exists). */
int
add_global_name(const char *name, struct type *data)
{
	if(find_name_index(1, name) >= 0) return 0;
	if(globalsz+1 >= gsize) enlarge_table(1);
	global_names[globalsz] = strdup(name);
	global_data[globalsz] = data;
	++globalsz;
	return 1;
}

int
add_local_name(const char *name, struct type *data)
{
	if(find_name_index(2, name) >= 0) return 0;
	if(localsz+1 >= lsize) enlarge_table(2);
	local_names[localsz] = strdup(name);
	local_data[localsz] = data;
	++localsz;
	return 1;
}

/*	Clear the symbol table. We free() only the items in local_names[] array.
 *	We can't free any of the items in local_data[] because these pointers are
 *	shared, and we can't free() them twice.
 */
void
clear_local_names(void)
{
	int i;

	for(i = 0; i < localsz; i++) free(local_names[i]);
	localsz = 0;
}

/*	Node allocation primitives. We need yylineno for error reporting during
 *	code generation phase. */
extern int yylineno;

/*	type is the type which we wish to make. Other parameters (if any; there are
 *	none for simple types) are dependent upon type. */
struct type *
new_type(int type, ...)
{
	int i;
	va_list ap;
	struct type *tmp = xmalloc(sizeof(struct type));

	tmp->type_tag = type;
	va_start(ap, type);

	switch(tmp->type_tag) {
	case ARRAY:		/* elemtype, ndims, dims... */
		tmp->class.array.elemtype = va_arg(ap, struct type*);
		tmp->class.array.ndims = va_arg(ap, int);
		if(tmp->class.array.ndims > 0) {
			tmp->class.array.dims[0] = va_arg(ap, int);
			if(tmp->class.array.ndims > 1) {
				tmp->class.array.dims[1] = va_arg(ap, int);
			}
			if(tmp->class.array.ndims > 2) FATAL("invalid array dimensions");
		}
		break;
	case PROCEDURE:	/* */
	case FUNCTION:	/* retarg */
		/*	this is called at the time when there are no local variables
		 *	except parameters. */
		tmp->class.subroutine.nargs = localsz;
		if(tmp->type_tag == FUNCTION) tmp->class.subroutine.retarg = va_arg(ap, struct type*);
		if(tmp->class.subroutine.nargs > 0) {
			tmp->class.subroutine.args = xmalloc(tmp->class.subroutine.nargs*sizeof(struct type*));
			for(i = 0; i < tmp->class.subroutine.nargs; i++) {
				tmp->class.subroutine.args[i] = local_data[i];
			}
		}
		break;
	case PROGRAM:	/* # args to dispatch on */
		tmp->class.subroutine.nargs = va_arg(ap, int);
		break;
	case IDN:	/* scope, pos */
		tmp->scope = va_arg(ap, int);
		tmp->pos = va_arg(ap, int);
		break;
	}

	va_end(ap);
	return tmp;
}

/*	lit_type is token type. Other parameter(s) are the constant value. */
struct st_node *
new_st_literal(int lit_type, ...)
{
	va_list ap;
	struct st_node *tmp = xmalloc(sizeof(struct st_node));

	tmp->opcode = lit_type;
	tmp->line = yylineno;
	va_start(ap, lit_type);

	switch(lit_type) {
		case LIT_REAL:
			tmp->type = new_type(REAL);
			tmp->class.literal.real = va_arg(ap, double);
			break;
		case LIT_NATURAL:
			tmp->type = new_type(NATURAL);
			tmp->class.literal.natural = va_arg(ap, int);
			break;
		case LIT_BOOLEAN:
			tmp->type = NULL;
			tmp->class.literal.boolean = va_arg(ap, int);
			break;
		case LIT_STRING:
			tmp->type = new_type(STRING);
			tmp->class.literal.str = strdup(va_arg(ap, char*));
			break;
		case LIT_BINARY:
			tmp->type = new_type(BINARY);
			tmp->class.literal.binary = strdup(va_arg(ap, char*));
			break;
		case LIT_CHAR:
			tmp->type = new_type(CHAR);
			tmp->class.literal.chr = va_arg(ap, char);
			break;
		case LIT_COMPLEX:
			tmp->type = new_type(COMPLEX);
			tmp->class.literal.complex[0] = va_arg(ap, double);
			tmp->class.literal.complex[1] = va_arg(ap, double);
			break;
		default:
			FATAL("invalid literal type");
	}

	va_end(ap);
	return tmp;
}

/*	creates new identifier in the syntax tree. id is identifier name. */
struct st_node *
new_st_identifier(const char *id)
{
	struct st_node *tmp = xmalloc(sizeof(struct st_node));
	
	tmp->opcode = IDN;
	tmp->class.identifier.name = strdup(id);
	return tmp;
}

/*	opcode is encodes the operations. type is relevant only in expressions;
 *	in control statements can be set to NULL. The rest of the parameters are
 *	the child nodes of this node. This list must be terminated by NULL. */
struct st_node *
new_st_control(int opcode, struct type *type, ...)
{
	va_list ap;
	struct st_node *tmp = xmalloc(sizeof(struct st_node));
	struct st_node *child;
	int child_ctr = 0;
	
	tmp->opcode = opcode;
	tmp->type = type;
	tmp->line = yylineno;
	va_start(ap, type);

	do {
		child = va_arg(ap, struct st_node*);
		if(child_ctr >= 5) FATAL("children list too small");
		tmp->class.control.children[child_ctr++] = child;
	} while(child != NULL);

	va_end(ap);
	return tmp;
}

