/*
 * optimize.c - optimization step
 * (c) 2000 Zeljko Vrba (mordor@fly.srk.fer.hr)
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <config.h>
#include "node.h"
#include "hpc-y.h"

#ifdef __GNUC__
static char rcsid[] __attribute__((unused)) = "$Id: optimize.c,v 1.1 2000-01-08 09:24:43+01 zeljko Exp $";
#else
static char rcsid[] = "$Id: optimize.c,v 1.1 2000-01-08 09:24:43+01 zeljko Exp $";
#endif

/* 	List of optimizations:
 *	1.	Constants like -3. are compiled into % 3 NEG %CHS
 *		Optimized to % -3.
 *		Currently works only on % and C%. I'm lazy to write 2nd complement
 *		on arbitrary HEX strings represented as C strings.
 *	2.	Expressions like complex(2., 3.) are compiled into % 2 % 3  %%>C%
 *		Optimize to C% 2 3
 *
 *	TODO:
 *	1.	All constant expressions are evaluated, like 3 - 2.
 *		Compiled to # 3 # 2 #-, optimize to # 1.
 *	2.	Optimize mixing of binary and real numbers:
 *		real #a34 + 12.
 *		Compiled to HXS 3 A34 HXS>% % 12 +.
 *		Optimize to HXS 3 A34 % 12 bit#%+.
 *	3.	Casts on constants like real 3 are compiled into # 3 UNCOERCE.
 *		Optimize to % 3. I deliberately left this as a secondary optimization
 *		because you can type in in the source "3.".
 *	4,	Other optimizations that combine multiple operations into single.
 */

/*	Unary minus optimization. */
static void
opt_uminus(struct st_node *root)
{
	struct st_node *child = root->class.control.children[0];

	switch(child->opcode) {
	case LIT_REAL:
		root->opcode = LIT_REAL;
		root->class.literal.real = -child->class.literal.real;
		break;
	case LIT_COMPLEX:
		root->opcode = LIT_COMPLEX;
		root->class.literal.complex[0] = -child->class.literal.complex[0];
		root->class.literal.complex[1] = -child->class.literal.complex[1];
		break;
	}
}

/*	RE2COMPLEX optimization. */
static void
opt_re2complex(struct st_node *root)
{
	struct st_node *c0 = root->class.control.children[0],
				   *c1 = root->class.control.children[1];
	
	if(c0->opcode == LIT_REAL && c1->opcode == LIT_REAL) {
		root->opcode = LIT_COMPLEX;
		root->class.literal.complex[0] = c0->class.literal.real;
		root->class.literal.complex[1] = c1->class.literal.real;
	}
}

/*	Main optimization function. */
void
optimize_code(struct st_node *root)
{
	int i;
	
	/*	null node and nodes that can't have any children... */
	if(!root) return;
	if(root->opcode == LIT_NATURAL || root->opcode == LIT_REAL
		|| root->opcode == LIT_BINARY || root->opcode == LIT_CHAR
		|| root->opcode == LIT_STRING || root->opcode == IDN
		|| root->opcode == KW_NULL
		|| root->opcode == PROGRAM || root->opcode == DISPATCH)
		return;
		

	/*	first optimize all children of this node. */
	for(i = 0; root->class.control.children[i] != 0; i++) {
		optimize_code(root->class.control.children[i]);
	}

	/*	optimize this particular node with alredy optimized children. */
	switch(root->opcode) {
	case UMINUS:
		opt_uminus(root);
		break;
	case RE2COMPLEX:
		opt_re2complex(root);
		break;
	}
}

