/*
 * saturn.c - writing executable code
 * (c) 2000 Zeljko Vrba (mordor@fly.srk.fer.hr)
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <config.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include "saturn.h"
#include "node.h"

#ifdef __GNUC__
static char rcsid[] __attribute__((unused)) = "$Id: saturn.c,v 1.1 2000-01-08 09:24:43+01 zeljko Exp $";
#else
static char rcsid[] = "$Id: saturn.c,v 1.1 2000-01-08 09:24:43+01 zeljko Exp $";
#endif

#ifndef HAVE_ECVT
char *ecvt(double, size_t, int*, int*);
#endif

#ifndef HAVE_STRDUP
char *strdup(char *);
#endif

/*	Hash table for entry points. */
#define	HSIZE	8191
struct hentry {
	char *name;
	int addr;
	struct hentry *next;
};

static struct hentry *hash[HSIZE];
static int hash_collisions = 0;
static int loaded = 0;

/*	Calculating the hash code of the string. */
static int
hash_code(const char *str)
{
	int hash = 0;

	for(; *str; str++) hash = ((hash * 613) + (long)*str) % HSIZE;
	return hash;
}

/*	Insertion. */
static void
hash_insert(const char *name, int addr)
{
	struct hentry *h = malloc(sizeof(struct hentry));
	int hindex = hash_code(name);

	if(hash[hindex] != NULL) ++hash_collisions;
	if(!h) {
		fputs("virtual memory exhausted\n", stderr);
		exit(1);
	}
	h->name = strdup(name);
	if(!h->name) {
		fputs("virtual memory exhausted\n", stderr);
		exit(1);
	}
	h->addr = addr;
	h->next = hash[hindex];
	hash[hindex] = h;
}

/*	Finding. Returns the address of the entry point. */
static int
hash_find(const char *name)
{
	struct hentry *h;
	int hindex = hash_code(name);

	h = hash[hindex];
	while(h) {
		if(strcmp(h->name, name) == 0) return h->addr;
		h = h->next;
	}
	fprintf(stderr, "error: unknown entry point: %s\n", name);
	exit(1);
}

static void
load_entries(const char *entries)
{
	FILE *f;
	char buf[1024];
	char *ent, *equ;
	int addr;

	if(!(f = fopen(entries, "r"))) {
		fprintf(stderr, "can't open entries file for reading\n");
		exit(1);
	}

	while(fgets(buf, 1024, f)) {
		ent = strtok(buf, " \t");
		if(ent[0] != '=') continue;
		equ = strtok(NULL, " \t");
		if(strcmp(equ, "EQU") != 0 && strcmp(equ, "equ") != 0) continue;
		equ = strtok(NULL, " \t");
		if(equ[0] == '#') {
			addr = strtol(equ+1, NULL, 16);
		} else {
			addr = strtol(equ+1, NULL, 10);
		}
		hash_insert(ent+1, addr);	/* +1 discards leading = */
	}
	fclose(f);
}

/*	This holds the name of the output file. */
static const char *outname;

/*	This is the buffer which holds nybbles, its allocated size and the
 *	number of nybbles actually in the buffer. */
static char *data;
static int dsize, nnyb;

static void
enlarge_buffer(int newsize)
{
	if(!(data = realloc(data, newsize))) {
		fputs("virtual memory exhausted\n", stderr);
		exit(1);
	}
	dsize = newsize;
}

/*	Writes a single nybble to the buffer. */
static void
nybble(int n)
{
	n &= 0xF;
	if(nnyb >> 1  >= dsize) enlarge_buffer(dsize * 2);
	if(nnyb & 1) {
		data[nnyb >> 1] |= (n << 4);
	} else {
		data[nnyb >> 1] |= n;
	}
	++nnyb;
}

/*	Opens a file for writing. The file can immediately be downloaded to the HP.
 *	Automatically writes download header. Reurns < 0 if unsuccessful. out is
 *	the output file name. entries is the name of the ASCII entries file. */
void
saturn_openw(const char *out, const char *entries)
{
	data = NULL; 
	outname = out;
	enlarge_buffer(1024);
	nnyb = 0;
	if(!loaded) {
		load_entries(entries);
		loaded = 1;
	}
}

/*	Writes a buffer to the file. Returns < 0 in case of any error. */
int
saturn_close(void)
{
	FILE *f;
	int result = 0;
	
	if(nnyb & 1) nybble(0);
	if(!(f = fopen(outname, "w"))) return -1;
	fputs("HPHP48-R", f);
	if(fwrite(data, 1, nnyb >> 1, f) < (nnyb >> 1)) result = -1;
	if(fclose(f) < 0) result = -1;
	return result;
}

/*	5-nybble pointer or binary integer. */
void
saturn_ptr5(int ptr)
{
	int i;

	for(i = 0; i < 5; i++) {
		nybble(ptr & 0xF);
		ptr >>= 4;
	}
}

/*	real (or extended real) number. Integer part and "." is mandatory.
 *	(defined in lexer). This code should be checked! */
void
saturn_real(double num, int extended)
{
	int i;
	int exponent, sign;
	char *mantissa;

	mantissa = ecvt(num, extended ? 15 : 12, &exponent, &sign);
	--exponent;

	/*	Output the normalized BCD. */
	if(exponent < 0) exponent = (extended ? 100000 : 1000) + exponent;
	for(i = 0; i < (extended ? 5 : 3); i++, exponent /= 10) nybble(exponent % 10);
	for(i = extended ? 14 : 11; i >= 0; i--) {
		nybble(mantissa[i] - '0');
	}
	nybble(sign ? 9 : 0);
}

/*	string. */
void
saturn_string(const char *str)
{
	int i, sl = strlen(str), sl1;
	
	sl1 = (sl << 1) + 5;
	for(i = 0; i < 5; i++) {
		nybble(sl1 & 0xF);
		sl1 >>= 4;
	}

	for(i = 0; i < sl; i++) {
		nybble(str[i] & 0xF);
		nybble(str[i] >> 4);
	}
}

/*	hex string. */
void
saturn_hexstr(int len, const char *hs)
{
	int i;
	
	saturn_ptr5(len);
	for(i = 0; i < len; i++) {
		nybble(hs[i] > '9' ? hs[i] - 55 : hs[i] - 48);
	}
}

/*	local/global name. */
void
saturn_id(const char *id)
{
	int i, sl = strlen(id);

	nybble(sl & 0xF); nybble(sl >> 4);
	for(i = 0; i < sl; i++) {
		nybble(id[i] & 0xF);
		nybble(id[i] >> 4);
	}
}

/*	named entry point. */
void
saturn_entry(const char *ent)
{
	int addr = hash_find(ent);
	saturn_ptr5(addr);
}

/*	This one is used to parse the printf-like string and to output appropriate
 *	SystemRPL code. This code requires that there are no more parameters after
 *	format; for example, in "PTR %x SEMI", SEMI will be ignored. */
void
saturn_print(const char *fmt, va_list ap)
{
	char buf[1024], *tok;

	strcpy(buf, fmt);
	for(tok = strtok(buf, " \t"); tok != NULL; tok = strtok(NULL, " \t")) {
		if(tok[0] == '*') return;	/* comment */
		if(strcmp(tok, "::") == 0) { saturn_entry("DOCOL"); continue; }
		if(strcmp(tok, ";") == 0) { saturn_entry("SEMI"); continue; }
		if(tok[0] == '{') { saturn_entry("DOLIST"); continue; }
		if(tok[0] == '}') { saturn_entry("SEMI"); continue; }
		if(strcmp(tok, "C%%") == 0) {
			saturn_entry("DOCMP");
			saturn_real(va_arg(ap, double), 0);
			saturn_real(va_arg(ap, double), 0);
		}
		if(strcmp(tok, "C%%%%") == 0) {
			saturn_entry("DOECMP");
			saturn_real(va_arg(ap, double), 1);
			saturn_real(va_arg(ap, double), 1);
			break;
		}
		if(strcmp(tok, "LAM") == 0) {
			saturn_entry("DOLAM");
			saturn_id(va_arg(ap, char*));
			break;
		}
		if(strcmp(tok, "CON(5)") == 0) {
			saturn_ptr5(va_arg(ap, int));
			break;
		}
		if(strcmp(tok, "ID") == 0) {
			saturn_entry("DOIDNT");
			saturn_id(va_arg(ap, char*));
			break;
		}
		if(strcmp(tok, "#") == 0) {
			saturn_entry("DOBINT");
			saturn_ptr5(va_arg(ap, int));
			break;
		}
		if(strcmp(tok, "%%") == 0) {
			saturn_entry("DOREAL");
			saturn_real(va_arg(ap, double), 0);
			break;
		}
		if(strcmp(tok, "%%%%") == 0) {
			saturn_entry("DOEREL");
			saturn_real(va_arg(ap, double), 1);
			break;
		}
		if(strcmp(tok, "HXS") == 0) {
			/*	note that we can't use
			 *	saturn_hexstr(va_arg(ap, int), va_arg(ap, char*))
			 *	since va_arg is a macro with a side-effect on ap */
			int len = va_arg(ap, int);
			char *str = va_arg(ap, char*);
			
			saturn_entry("DOHXS");
			saturn_hexstr(len, str);
			break;
		}
		if(strcmp(tok, "CHR") == 0) {
			char c = va_arg(ap, char);
			saturn_entry("DOCHAR");
			nybble(c & 0xF); nybble(c >> 4);
			break;
		}
		if(strcmp(tok, "$") == 0) {
			saturn_entry("DOCSTR");
			saturn_string(va_arg(ap, char*));
			break;
		}
		saturn_entry(tok);
	}
}

